C         ************************************************
C         ***    SUBROUTINE -- TRIANG                  ***
C         ************************************************

          SUBROUTINE TRIANG

          INCLUDE 'triang.cmn'

C         begin Delunay triangulation
                                 
               CALL INVERT_CELL(CELL,RCELL)
               CALL INITIALIZE_BOX

               CALL SAVE_POSITION

3              CONTINUE

               CALL INITIALIZE_TRIANG
1              CONTINUE

               CALL FIRST_SIMPLEX

2              CONTINUE

               IF (DEGENERATION) THEN
                  CALL RANDOM_DISPLACE
                  GOTO 3
               ENDIF 

               CALL FIND_OPEN_FACE
                              
               IF (OPEN_FACE_EXIST) THEN
                  CALL NEXT_SIMPLEX
                  GOTO 2
               ENDIF     
  
               CALL FIND_OPEN_CENTER

               IF (OPEN_CENTER_EXIST) GOTO 1                          

               CALL RESTORE_POSITION

               CALL SIMPL_TETR

C         Delunay triangulation complete             

          END         
                
C         *****************************************************
C         ***      FIRST LEVEL                              ***
C         ***      SUBROUTINES -- INITIALIZE_TRIANG         *** 
C         ***                  -- FIRST_SIMPLEX             ***
C         ***                  -- FIND_OPEN_FACE            ***
C         ***                  -- NEXT_SIMPLEX              ***
C         *****************************************************
          
          SUBROUTINE INITIALIZE_BOX

          INCLUDE 'triang.cmn'

             BX=CELL(1)**2+CELL(2)**2+CELL(3)**2
             BX=DSQRT(BX)
             BY=CELL(4)**2+CELL(5)**2+CELL(6)**2
             BY=DSQRT(BY)
             BZ=CELL(7)**2+CELL(8)**2+CELL(9)**2
             BZ=DSQRT(BZ)

          END

C**************************************************************

          SUBROUTINE INVERT_CELL(A,B)
          
          DOUBLE PRECISION A,B,D,R                                        
          DIMENSION A(9),B(9)

c     
c         calculate adjoint matrix

          b(1)=a(5)*a(9)-a(6)*a(8)
          b(2)=a(3)*a(8)-a(2)*a(9)
          b(3)=a(2)*a(6)-a(3)*a(5)
          b(4)=a(6)*a(7)-a(4)*a(9)
          b(5)=a(1)*a(9)-a(3)*a(7)
          b(6)=a(3)*a(4)-a(1)*a(6)
          b(7)=a(4)*a(8)-a(5)*a(7)
          b(8)=a(2)*a(7)-a(1)*a(8)
          b(9)=a(1)*a(5)-a(2)*a(4)
c     
c         calculate determinant

          d=a(1)*b(1)+a(4)*b(2)+a(7)*b(3)
          r=0.d0
          if(abs(d).gt.0.d0)r=1.d0/d
c     
c         complete inverse matrix

          b(1)=r*b(1)
          b(2)=r*b(2)
          b(3)=r*b(3)
          b(4)=r*b(4)
          b(5)=r*b(5)
          b(6)=r*b(6)
          b(7)=r*b(7)
          b(8)=r*b(8)
          b(9)=r*b(9)

          END
          
C********************************************** 

          SUBROUTINE IMAGES(X,Y,Z,CELL,RCELL)
                                               
          DOUBLE PRECISION X,Y,Z,XN,YN,ZN
          DOUBLE PRECISION CELL(9),RCELL(9)
          DOUBLE PRECISION SSX,SSY,SSZ
          DOUBLE PRECISION XSS,YSS,ZSS

          ssx=(rcell(1)*x+rcell(4)*y+rcell(7)*z)
          ssy=(rcell(2)*x+rcell(5)*y+rcell(8)*z)
          ssz=(rcell(3)*x+rcell(6)*y+rcell(9)*z)
          
          xss=ssx-dnint(ssx)
          yss=ssy-dnint(ssy)
          zss=ssz-dnint(ssz)
          
          x=(cell(1)*xss+cell(4)*yss+cell(7)*zss)
          y=(cell(2)*xss+cell(5)*yss+cell(8)*zss)
          z=(cell(3)*xss+cell(6)*yss+cell(9)*zss)


          END
          
C**********************************************          
          
          SUBROUTINE INITIALIZE_TRIANG

          INCLUDE 'triang.cmn'

               IOPC=ICENT(1)          
               IS=0
               IE=0

               IF (NN.LT.40) THEN
                  MIN_NUM_CAN=1
               ELSE   
                  MIN_NUM_CAN=MIN0(200,(NN-4)/2)
               ENDIF  
               MAX_NUM_CAN=MIN0(300,2*NN)

               CUT_OFF=MAX_NUM_CAN*BX*BY*BZ/(NN*4D0)
               CUT_OFF=CUT_OFF**0.33333333
               DEGENERATION=.FALSE.
               WARNING_CUT=.FALSE.

               DO I=1,NC
                  IN_SIMP(I)=.FALSE. 
               ENDDO

               HALF_BOX=DMAX1(BX,BY,BZ)

               DISP_PAR=DMIN1(BX,BY,BZ)
               DISP_PAR=DISP_PAR**2 

               DO I=1,1
                  XI=CENT(1,I)
                  YI=CENT(2,I)
                  ZI=CENT(3,I)
                  DO J=1,NN
                     IF (ICENT(I).NE.INEIG(J)) THEN
                        XJ=NEIG(1,J)
                        YJ=NEIG(2,J)
                        ZJ=NEIG(3,J)
                        X=XI-XJ
                        Y=YI-YJ
                        Z=ZI-ZJ
                        CALL IMAGES(X,Y,Z,CELL,RCELL)
                        D=X**2+Y**2+Z**2
                        DISP_PAR=DMIN1(DISP_PAR,D)
                     ENDIF
                  ENDDO
               ENDDO

               DISP_PAR=SQRT(DISP_PAR)

               IF (DISP_PAR.EQ.0D0)
     :            CALL TRIANG_ERROR_MESSAGE
     :            ('disp_par is equal 0',' ')

               MIN_DOUBLE=DISP_PAR*1D-7
               MAX_DOUBLE=DISP_PAR*1D20
               DISP_PAR=DISP_PAR*1D-5

          END

C         *** END INITIALIZE_TRIANG **********************
  
          SUBROUTINE FIRST_SIMPLEX

          INCLUDE 'triang.cmn'
           
               IS=IS+1     
               
               IF (IS.GT.MAX_NUM_SIMP) 
     :             CALL TRIANG_ERROR_MESSAGE
     :             ('-IS- is greater then MAX_NUM_SIMP in',
     :              ' FIRST_SIMPLEX !')
      
               A(1)=IOPC
               INI_SIMP=.TRUE.

               CALL FIND_2

               CALL FIND_3

               CALL FIND_4
               
          END

C         *** END FIRST_SIMPLEX ***********************

          SUBROUTINE FIND_OPEN_FACE

          INCLUDE 'triang.cmn'

              OPEN_FACE_EXIST=.TRUE. 
              DO I=1,IS
                 IF (OPEN_SIMP(I)) THEN
                    DO J=1,4
                       IF (OP_FACE(J,I)) THEN
                          IOPS=I
                          IOPF=J
                          GOTO 1
                       ENDIF
                    ENDDO
                 ENDIF
              ENDDO
              
              OPEN_FACE_EXIST=.FALSE. 
1             CONTINUE
  
          END

C         *** END FIND_OPEN_FACES **********************

          SUBROUTINE NEXT_SIMPLEX

          INCLUDE 'triang.cmn'
          INTEGER A1,A2,A3,A4
          DOUBLE PRECISION IL4,ILI
          INTEGER NAI(MAX_NUM_NEIG),DDC(100)
          DOUBLE PRECISION T(MAX_NUM_NEIG) 
   
               J=0
               DO I=1,4
                  IF (I.NE.IOPF) THEN 
                     J=J+1
                     A(J)=SIMP(I,IOPS)
                  ENDIF
               ENDDO
               A(4)=SIMP(IOPF,IOPS)

               IF (IN_CENT(A(1))) THEN
                  A1=IND_CENT(A(1))
                  RX1=CENT(1,A1)                
                  RY1=CENT(2,A1)                
                  RZ1=CENT(3,A1) 
                  RAD1=CENT(4,A1) 
               ELSE
                  A1=IND_NEIG(A(1))
                  RX1=NEIG(1,A1)                
                  RY1=NEIG(2,A1)                
                  RZ1=NEIG(3,A1) 
                  RAD1=NEIG(4,A1)
               ENDIF
                        
               IF (IN_CENT(A(2))) THEN
                  A2=IND_CENT(A(2))
                  RX2=CENT(1,A2)                
                  RY2=CENT(2,A2)                
                  RZ2=CENT(3,A2)
                  RAD2=CENT(4,A2) 
               ELSE
                  A2=IND_NEIG(A(2))
                  RX2=NEIG(1,A2)                
                  RY2=NEIG(2,A2)                
                  RZ2=NEIG(3,A2)
                  RAD2=NEIG(4,A2)
               ENDIF
               
               IF (IN_CENT(A(3))) THEN
                  A3=IND_CENT(A(3))
                  RX3=CENT(1,A3)                
                  RY3=CENT(2,A3)                
                  RZ3=CENT(3,A3)  
                  RAD3=CENT(4,A3) 
               ELSE
                  A3=IND_NEIG(A(3))
                  RX3=NEIG(1,A3)                
                  RY3=NEIG(2,A3)                
                  RZ3=NEIG(3,A3)     
                  RAD3=NEIG(4,A3) 
               ENDIF   
               
               IF (IN_CENT(A(4))) THEN
                  A4=IND_CENT(A(4)) 
                  RX4=CENT(1,A4)                
                  RY4=CENT(2,A4)                
                  RZ4=CENT(3,A4)
               ELSE
                  A4=IND_NEIG(A(4)) 
                  RX4=NEIG(1,A4)                
                  RY4=NEIG(2,A4)                
                  RZ4=NEIG(3,A4)
               ENDIF

               X2=RX2-RX1
               Y2=RY2-RY1
               Z2=RZ2-RZ1
               CALL IMAGES(X2,Y2,Z2,CELL,RCELL)
               
               X3=RX3-RX1               
               Y3=RY3-RY1
               Z3=RZ3-RZ1
               CALL IMAGES(X3,Y3,Z3,CELL,RCELL)
               
               X4=RX4-RX1
               Y4=RY4-RY1
               Z4=RZ4-RZ1
               CALL IMAGES(X4,Y4,Z4,CELL,RCELL)
               
               AX=Y2*Z3-Y3*Z2
               AY=X3*Z2-X2*Z3
               AZ=X2*Y3-Y2*X3
               DA=AX**2+AY**2+AZ**2
               DA=1D0/DSQRT(DA)
               AX=AX*DA 
               AY=AY*DA
               AZ=AZ*DA

               IL4=AX*X4+AY*Y4+AZ*Z4
               IF (IL4.GT.0D0) THEN
                  AX=-AX
                  AY=-AY
                  AZ=-AZ
               ENDIF
                             
               R12=X2**2+Y2**2+Z2**2  
               SQ12=DSQRT(R12)
               D12=RAD1**2-RAD2**2+R12
               D12=D12/SQ12
               D12=D12*0.5D0

               R13=X3**2+Y3**2+Z3**2
               SQ13=DSQRT(R13)
               D13=RAD1**2-RAD3**2+R13
               D13=D13/SQ13
               D13=D13*0.5D0
               
               COSFI=X3*X2+Y3*Y2+Z3*Z2
               COSFI=COSFI/(SQ12*SQ13)
               
               B2=(1D0-(COSFI*D12/D13))
               B2=B2/(1D0-COSFI**2)
               B1=1D0-(B2*COSFI*D13/D12)
              
               BIX=B1*X2*D12/SQ12+B2*X3*D13/SQ13
               BIY=B1*Y2*D12/SQ12+B2*Y3*D13/SQ13
               BIZ=B1*Z2*D12/SQ12+B2*Z3*D13/SQ13
               
1              CONTINUE

               TMIN=MAX_DOUBLE                         
               ICAN=0  


               DO I=1,NN
                  J=INEIG(I)
                  IF ((J.NE.A(1)).AND.
     :                (J.NE.A(2)).AND.
     :                (J.NE.A(3))) THEN
     :                 
                      RADI=NEIG(4,I)       
                      RAD=RAD1+RADI
                      XI=NEIG(1,I)-RX1
                      YI=NEIG(2,I)-RY1
                      ZI=NEIG(3,I)-RZ1                      

                      CALL IMAGES(XI,YI,ZI,CELL,RCELL)                      

                      IF ((DABS(XI)-RAD).LT.CUT_OFF) THEN
                         IF ((DABS(YI)-RAD).LT.CUT_OFF) THEN
                            IF ((DABS(ZI)-RAD).LT.CUT_OFF) THEN

                               ILI=AX*XI+AY*YI+AZ*ZI

                               IF (ILI.GT.0D0) THEN
                                  
                                  ICAN=ICAN+1
                                 
                                  R1I=XI**2+YI**2+ZI**2
                                  D1I=RAD1**2-RADI**2+R1I      
                                  
                                  IF (D1I.LT.0D0) then 
                                     CALL TRIANG_WARNING_MESSAGE
     :                               ('ball -1- inside -i-',' ')
                                     CALL WRITE_NUMBERS
     :                               (A(1),INEIG(I))
                                  ENDIF
                                  
                                   D1I=D1I/(2D0*R1I)
                               
                                   DIX=XI*D1I
                                   DIY=YI*D1I
                                   DIZ=ZI*D1I
                                   T1=XI*(BIX-DIX)+
     :                                YI*(BIY-DIY)+
     :                                ZI*(BIZ-DIZ)
                                   T2=XI*AX+
     :                                YI*AY+
     :                                ZI*AZ                   

                                  IF (DABS(T2).LT.
     :                                MIN_DOUBLE) THEN

                                      T(ICAN)=MAX_DOUBLE

                                  ELSE
                                
                                      T(ICAN)=-T1/T2

                                  ENDIF 

                                  NAI(ICAN)=I 
                                  TMIN=DMIN1(T(ICAN),TMIN)
  
                               ENDIF

                            ENDIF
                         ENDIF
                      ENDIF
                  ENDIF  
               ENDDO  

               IF (ICAN.LT.MIN_NUM_CAN) THEN  
               
                  IF (CUT_OFF.GT.HALF_BOX) THEN
                  
                     IF (.NOT.WARNING_CUT) THEN
                        WARNING_CUT=.TRUE.
                        CALL TRIANG_WARNING_MESSAGE
     :          ('CUT_OFF is greater then HALF_BOX in NEXT_SIMPLEX',
     :           ', but ICAN is lower then MIN_NUM_CAN')
                     ENDIF
    
                     IF (ICAN.EQ.0) THEN
                        CALL TRIANG_ERROR_MESSAGE
     :             ('new atom in NEXT_SIMPLEX not found',' ')
                     ENDIF 
                   
                    GOTO 111
                       
                  ENDIF
                  
                  CUT_OFF=CUT_OFF/CUT_OFF_FACTOR
                  GOTO 1
                  
               ENDIF   

111            CONTINUE
                   
               IF (ICAN.GT.MAX_NUM_CAN) 
     :            CUT_OFF=CUT_OFF*CUT_OFF_FACTOR          

               OP_FACE(IOPF,IOPS)=.FALSE.

               J=0
               DO I=1,ICAN
                  IF (DABS(T(I)-TMIN).LT.MIN_DOUBLE) THEN
                     J=J+1
                     IF (J.GT.100) 
     :                  CALL TRIANG_ERROR_MESSAGE
     :                  ('number of degenerated in NEXT_SIMPLEX ',
     :                   ' is too big')
                     DDC(J)=NAI(I)
                   ENDIF
               ENDDO

               IF (J.EQ.0) 
     :             CALL TRIANG_ERROR_MESSAGE
     :            ('size of ddc is zero in NEXT_SIMPLEX',
     :             ' !') 

               IS=IS+1            

               IF (IS.GT.MAX_NUM_SIMP) 
     :            CALL TRIANG_ERROR_MESSAGE
     :            ('-IS- is greater then MAX_NUM_SIMP in',
     :             ' NEXT_SIMPLEX !') 
 
               X=AX*TMIN+BIX+RX1
               Y=AY*TMIN+BIY+RY1
               Z=AZ*TMIN+BIZ+RZ1
               
               CALL IMAGES(X,Y,Z,CELL,RCELL)
               
               VOR_VER(1,IS)=X               
               VOR_VER(2,IS)=Y               
               VOR_VER(3,IS)=Z

               A(4)=INEIG(DDC(1))

               INI_SIMP=.FALSE.
              
               CALL ADD_SIMPLEX
               CALL OPEN_FACES

               IF (J.GT.1)  DEGENERATION=.TRUE.                           

          END

C         *** END NEXT_SIMPLEX ***************************

          SUBROUTINE FIND_OPEN_CENTER

          INCLUDE 'triang.cmn'

              OPEN_CENTER_EXIST=.FALSE.
              DO I=1,IS
                 DO J=1,4
                    K=SIMP(J,I)                 
                    IF (IN_CENT(K)) IN_SIMP(IND_CENT(K))=.TRUE.
                 ENDDO
              ENDDO
              
              DO I=1,NC
                 IF (.NOT.IN_SIMP(I)) THEN
                    OPEN_CENTER_EXIST=.TRUE. 
                    IOPC=ICENT(I)
                    GOTO 1
                 ENDIF
              ENDDO
              
1             CONTINUE
  
          END                              
          
C         *** END FIND_OPEN_CENTER ***********************

C         *****************************************************
C         ***      SECOND LEVEL                             ***
C         ***      SUBROUTINES -- FIND_2                    *** 
C         ***                  -- FIND_3                    ***
C         ***                  -- FIND_4                    ***
C         ***                  -- OPEN_FACES                ***
C         ***                  -- TRIANG_ERROR_MESSAGE      ***  
C         ***                  -- TRIANG_WARNING_MESSAGE    ***
C         ***                  -- ADD_SIMPLEX               ***
C         *****************************************************

          SUBROUTINE FIND_2

          INCLUDE 'triang.cmn'
          INTEGER A1,A2             
          DOUBLE PRECISION MAX_D1I

               A1=IND_CENT(A(1))
               RX1=CENT(1,A1)                
               RY1=CENT(2,A1)                
               RZ1=CENT(3,A1) 
               RAD1=CENT(4,A1) 
               
1              CONTINUE
               
               ICAN=0         
               MAX_D1I=MAX_DOUBLE
               
               DO I=1,NN        
                  IF (A(1).NE.INEIG(I)) THEN
                      
                      RADI=NEIG(4,I)
                      RAD=RAD1+RADI
                      XI=NEIG(1,I)-RX1    
                      YI=NEIG(2,I)-RY1 
                      ZI=NEIG(3,I)-RZ1
                      
                      CALL IMAGES(XI,YI,ZI,CELL,RCELL)
                     
                      IF ((DABS(XI)-RAD).LT.CUT_OFF) THEN                         
                         IF ((DABS(YI)-RAD).LT.CUT_OFF) THEN                                        
                            IF ((DABS(ZI)-RAD).LT.CUT_OFF) THEN

                                ICAN=ICAN+1
                                  
                                R1I=XI**2+YI**2+ZI**2
                                D1I=RAD1**2-RADI**2+R1I  
                                
                                IF (D1I.LT.0D0) THEN 
                                   CALL TRIANG_WARNING_MESSAGE
     :                             ('ball -1- inside -i-',' ')
                                   CALL WRITE_NUMBERS
     :                             (A(1),INEIG(I))
                                ENDIF          

                                D1I=D1I/DSQRT(R1I)
                                
                                IF (D1I.LT.MAX_D1I) THEN
                                   MAX_D1I=D1I
                                   A2=I       
                                ENDIF        
                                
                            ENDIF       
                         ENDIF
                      ENDIF         
                  ENDIF         
               ENDDO   

               IF (ICAN.LT.MIN_NUM_CAN) THEN  
               
                  IF (CUT_OFF.GT.HALF_BOX) THEN
                  
                     IF (.NOT.WARNING_CUT) THEN
                        WARNING_CUT=.TRUE.
                        CALL TRIANG_WARNING_MESSAGE
     :          ('CUT_OFF is greater then HALF_BOX in FIND_2',
     :           ', but ICAN is lower then MIN_NUM_CAN')               
                     ENDIF
   
                     IF (ICAN.EQ.0) THEN
                        CALL TRIANG_ERROR_MESSAGE
     :             ('new atom in FIND_2 not found','!')   
                     ENDIF
                        
                    GOTO 111
                       
                  ENDIF
                  
                  CUT_OFF=CUT_OFF/CUT_OFF_FACTOR
                  GOTO 1
                  
               ENDIF                      

111            CONTINUE
               
               IF (ICAN.GT.MAX_NUM_CAN) 
     :            CUT_OFF=CUT_OFF*CUT_OFF_FACTOR    
     
               A(2)=INEIG(A2)    
               
          END
          
C         *** END FIND_2 ***********************************
          
          SUBROUTINE FIND_3

          INCLUDE 'triang.cmn'
          INTEGER A1,A2,A3             
          DOUBLE PRECISION MAX_DD

               A1=IND_CENT(A(1))
               RX1=CENT(1,A1)                
               RY1=CENT(2,A1)                
               RZ1=CENT(3,A1) 
               RAD1=CENT(4,A1)  
               
               A2=IND_NEIG(A(2))
               RX2=NEIG(1,A2)                
               RY2=NEIG(2,A2)                
               RZ2=NEIG(3,A2) 
               RAD2=NEIG(4,A2)  
               
               X2=RX2-RX1          
               Y2=RY2-RY1               
               Z2=RZ2-RZ1              
               
               CALL IMAGES(X2,Y2,Z2,CELL,RCELL)
              
               R12=X2**2+Y2**2+Z2**2
               D12=RAD1**2-RAD2**2+R12
               D12=D12/DSQRT(R12)
               D12=D12*0.5D0
               
1              CONTINUE
               
               ICAN=0           
               MAX_DD=MAX_DOUBLE
               
               DO I=1,NN 
                  J=INEIG(I)
                  IF ((A(1).NE.J).AND.
     :                (A(2).NE.J)) THEN
                      
                      RADI=NEIG(4,I)
                      RAD=RAD1+RADI
                      XI=NEIG(1,I)-RX1                
                      YI=NEIG(2,I)-RY1
                      ZI=NEIG(3,I)-RZ1
                      
                      CALL IMAGES(XI,YI,ZI,CELL,RCELL)                      
                      
                      IF ((DABS(XI)-RAD).LT.CUT_OFF) THEN                         
                         IF ((DABS(YI)-RAD).LT.CUT_OFF) THEN                                        
                            IF ((DABS(ZI)-RAD).LT.CUT_OFF) THEN
                                         
                                ICAN=ICAN+1
                                  
                                R1I=XI**2+YI**2+ZI**2
                                D1I=RAD1**2-RADI**2+R1I

                                IF (D1I.LT.0D0) THEN 
                                   CALL TRIANG_WARNING_MESSAGE
     :                             ('ball -1- inside -i-',' ')
                                   CALL WRITE_NUMBERS
     :                             (A(1),INEIG(I))
                                ENDIF
                                
                                COSFI=X2*XI+Y2*YI+Z2*ZI
                                COSFI=COSFI/DSQRT(R12*R1I)
                                SCOS=DABS(1D0-COSFI**2)  

                                IF (SCOS.LT.
     :                                  MIN_DOUBLE) THEN

                                   DD=MAX_DOUBLE

                                ELSE

                                   D1I=D1I/DSQRT(R1I)
                                   D1I=0.5D0*D1I                                
                                   DD=((D1I/D12)-COSFI)**2
                                   DD=DD/SCOS
 
                                ENDIF            
                    
                                IF (DD.LT.MAX_DD) THEN
                                   MAX_DD=DD
                                   A3=I       
                                ENDIF        
                                
                            ENDIF       
                         ENDIF
                      ENDIF         
                  ENDIF         
               ENDDO   
               
               IF (ICAN.LT.MIN_NUM_CAN) THEN  
               
                  IF (CUT_OFF.GT.HALF_BOX) THEN
                  
                     IF (.NOT.WARNING_CUT) THEN
                        WARNING_CUT=.TRUE.
                       CALL TRIANG_WARNING_MESSAGE
     :          ('CUT_OFF is greater then HALF_BOX in FIND_3',
     :           ', but ICAN is lower then MIN_NUM_CAN')               
                     ENDIF

                     IF (ICAN.EQ.0)THEN 
                        CALL TRIANG_ERROR_MESSAGE
     :             ('new atom in FIND_3 not found','!')   
                     ENDIF
                       
                    GOTO 111
                       
                  ENDIF
                  
                  CUT_OFF=CUT_OFF/CUT_OFF_FACTOR
                  GOTO 1
                  
               ENDIF                      

111            CONTINUE
               
               IF (ICAN.GT.MAX_NUM_CAN) 
     :            CUT_OFF=CUT_OFF*CUT_OFF_FACTOR    
     
               A(3)=INEIG(A3)    
              
          END              
          
C         *** END FIND_3 **********************************

          SUBROUTINE FIND_4
          
          INCLUDE 'triang.cmn'
          INTEGER A1,A2,A3 
          INTEGER NAI(MAX_NUM_NEIG),DDC(100)
          DOUBLE PRECISION T(MAX_NUM_NEIG) 
               
               A1=IND_CENT(A(1))
               RX1=CENT(1,A1)                
               RY1=CENT(2,A1)                
               RZ1=CENT(3,A1) 
               RAD1=CENT(4,A1)
                        
               A2=IND_NEIG(A(2))
               RX2=NEIG(1,A2)                
               RY2=NEIG(2,A2)                
               RZ2=NEIG(3,A2)
               RAD2=NEIG(4,A2)

               A3=IND_NEIG(A(3))
               RX3=NEIG(1,A3)                
               RY3=NEIG(2,A3)                
               RZ3=NEIG(3,A3)     
               RAD3=NEIG(4,A3) 

               X2=RX2-RX1               
               Y2=RY2-RY1               
               Z2=RZ2-RZ1               
               CALL IMAGES(X2,Y2,Z2,CELL,RCELL)
               
               X3=RX3-RX1            
               Y3=RY3-RY1             
               Z3=RZ3-RZ1          
               CALL IMAGES(X3,Y3,Z3,CELL,RCELL)
               
               AX=Y2*Z3-Y3*Z2
               AY=X3*Z2-X2*Z3
               AZ=X2*Y3-Y2*X3
               DA=AX**2+AY**2+AZ**2
               DA=1D0/DSQRT(DA)
               AX=AX*DA 
               AY=AY*DA
               AZ=AZ*DA

               R12=X2**2+Y2**2+Z2**2  
               SQ12=DSQRT(R12)
               D12=RAD1**2-RAD2**2+R12
               D12=D12/SQ12
               D12=D12*0.5D0

               R13=X3**2+Y3**2+Z3**2
               SQ13=DSQRT(R13)
               D13=RAD1**2-RAD3**2+R13
               D13=D13/SQ13
               D13=D13*0.5D0
 
               COSFI=X3*X2+Y3*Y2+Z3*Z2
               COSFI=COSFI/(SQ12*SQ13)
               
               B2=(1D0-(COSFI*D12/D13))
               B2=B2/(1D0-COSFI**2)
               B1=1D0-(B2*COSFI*D13/D12)
              
               BIX=B1*X2*D12/SQ12+B2*X3*D13/SQ13
               BIY=B1*Y2*D12/SQ12+B2*Y3*D13/SQ13
               BIZ=B1*Z2*D12/SQ12+B2*Z3*D13/SQ13

1              CONTINUE

               TMIN=MAX_DOUBLE                         
               ICAN=0  

               DO I=1,NN
                  J=INEIG(I)
                  IF ((J.NE.A(1)).AND.
     :                (J.NE.A(2)).AND.
     :                (J.NE.A(3))) THEN
                             
                      RADI=NEIG(4,I)       
                      RAD=RAD1+RADI
                      XI=NEIG(1,I)-RX1 
                      YI=NEIG(2,I)-RY1 
                      ZI=NEIG(3,I)-RZ1
                      
                      CALL IMAGES(XI,YI,ZI,CELL,RCELL)
                       
                      IF ((DABS(XI)-RAD).LT.CUT_OFF) THEN                          
                         IF ((DABS(YI)-RAD).LT.CUT_OFF) THEN                             
                            IF ((DABS(ZI)-RAD).LT.CUT_OFF) THEN

                               T2=XI*AX+
     :                            YI*AY+
     :                            ZI*AZ                   

                               IF (T2.GT.MIN_DOUBLE) THEN
 
                                  ICAN=ICAN+1
                                  
                                  R1I=XI**2+YI**2+ZI**2
                                  D1I=RAD1**2-RADI**2+R1I
                               
                                  IF (D1I.LT.0D0) then 
                                     CALL TRIANG_WARNING_MESSAGE
     :                               ('ball -1- inside -i-',' ')
                                     CALL WRITE_NUMBERS
     :                               (A(1),INEIG(I))
                                  ENDIF
                         
                                  D1I=D1I/(2D0*R1I)
                                  DIX=XI*D1I
                                  DIY=YI*D1I
                                  DIZ=ZI*D1I
                                  T1=XI*(BIX-DIX)+
     :                               YI*(BIY-DIY)+
     :                               ZI*(BIZ-DIZ)

                                  T(ICAN)=DABS(T1/T2)
                                  NAI(ICAN)=I 
                                  TMIN=DMIN1(T(ICAN),TMIN)

                              ENDIF 
                            ENDIF
                         ENDIF
                      ENDIF
                  ENDIF  
               ENDDO  
               
               IF (ICAN.LT.MIN_NUM_CAN) THEN  
               
                  IF (CUT_OFF.GT.HALF_BOX) THEN
                  
                     IF (.NOT.WARNING_CUT) THEN
                        WARNING_CUT=.TRUE.
                        CALL TRIANG_WARNING_MESSAGE
     :          ('CUT_OFF is greater then HALF_BOX in FIND_4',
     :           ', but ICAN is lower then MIN_NUM_CAN')               
                     ENDIF
   
                     IF (ICAN.EQ.0) THEN
                         CALL TRIANG_ERROR_MESSAGE
     :             ('new atom in FIND_4 not found','!')   
                     ENDIF
                     
                    GOTO 111
                       
                  ENDIF
                  
                  CUT_OFF=CUT_OFF/CUT_OFF_FACTOR
                  GOTO 1
                  
               ENDIF   

111            CONTINUE
                   
               IF (ICAN.GT.MAX_NUM_CAN) 
     :            CUT_OFF=CUT_OFF*CUT_OFF_FACTOR 
     
               J=0
               DO I=1,ICAN
                  IF (DABS(T(I)-TMIN).LT.MIN_DOUBLE) THEN
                     J=J+1
                     IF (J.GT.100) 
     :                  CALL TRIANG_ERROR_MESSAGE
     :                  ('number of degenerated in FIND_4 ',
     :                   ' is too big')
                     DDC(J)=NAI(I)
                   ENDIF
               ENDDO

               IF (J.EQ.0) 
     :            CALL TRIANG_ERROR_MESSAGE
     :            ('size of ddc is zero in FIND_4',
     :             ' !') 
 
               X=AX*TMIN+BIX+RX1
               Y=AY*TMIN+BIY+RY1
               Z=AZ*TMIN+BIZ+RZ1
               
               CALL IMAGES(X,Y,Z,CELL,RCELL)
               
               VOR_VER(1,IS)=X               
               VOR_VER(2,IS)=Y               
               VOR_VER(3,IS)=Z
               
               
               A(4)=INEIG(DDC(1))
              
               CALL ADD_SIMPLEX
               CALL OPEN_FACES
                          
               IF (J.GT.1)  DEGENERATION=.TRUE. 
               

          END

C         *** END FIND_4 ***********************************

          SUBROUTINE OPEN_FACES

          INCLUDE 'triang.cmn'                         
          LOGICAL O_SIMP, FACE_EQUAL

               DO I=1,4
                  OP_FACE(I,IS)=.TRUE.
               ENDDO

               K=0
               KK=0
               DO I=1,4
                  J=SIMP(I,IS)               
                  IF (IN_CENT(J)) THEN 
                     K=K+1  
                     JJ=I
                     IF (.NOT.IN_NEIG(J)) KK=KK+1
                  ENDIF
               ENDDO
                       
               IF (K.EQ.0) CALL TRIANG_ERROR_MESSAGE
     :                     ('no centers in simplex detected',
     :                      ' in OPEN_FACES')

               IF (KK.GT.1) CALL TRIANG_ERROR_MESSAGE
     :                     ('more then qne pure centers is ',
     :                      'in simplex in OPEN_FACES')
     
               IF (K.EQ.1) THEN
                   OP_FACE(JJ,IS)=.FALSE.
               ENDIF

               IF (.NOT.INI_SIMP) THEN
                  
                   OP_FACE(NEW_AI,IS)=.FALSE.
                   DO I=1,IS-1                       
                      IF (OPEN_SIMP(I)) THEN
                         DO J=1,4
                            IF (OP_FACE(J,I)) THEN
                               DO K=1,4 
                               
                                  IF (OP_FACE(K,IS)) THEN
                                     
                                      IF (FACE_EQUAL(SIMP(1,I),J,
     :                                     SIMP(1,IS),K))THEN
                                          OP_FACE(K,IS)=.FALSE.
                                          OP_FACE(J,I)=.FALSE.
                                      ENDIF
                                          
                                  ENDIF    
                                                 
                               ENDDO
                            ENDIF
                         ENDDO
                      ENDIF
                   ENDDO
                 
                   DO I=1,IS                        
                      O_SIMP=.FALSE.     
                      DO J=1,4
                         O_SIMP=O_SIMP.OR.OP_FACE(J,I)
                      ENDDO         
                      IF (.NOT.O_SIMP) OPEN_SIMP(I)=.FALSE.
                   ENDDO
                 
               ENDIF
                              
          END

C         *** END OPEN_FACES **********************************

          SUBROUTINE TRIANG_ERROR_MESSAGE(TEXT1,TEXT2)

              CHARACTER  TEXT1*(*),TEXT2*(*)
              WRITE(*,'(A)') ' TRIANG FATAL ERROR:'
              WRITE(*,'(3A)') ' ',TEXT1,TEXT2
              WRITE(*,*) 
              STOP
 
          END                       
          
C         *** END TRIANG_ERROR_MESSAGE ********************

          SUBROUTINE TRIANG_WARNING_MESSAGE(TEXT1,TEXT2)
          
              CHARACTER  TEXT1*(*),TEXT2*(*)
              WRITE(*,'(A)') ' TRIANG WARNING:'
              WRITE(*,'(3A)') ' ',TEXT1,TEXT2
              WRITE(*,*)
              
          END                        
          
C         *** END TRIANG_WARNING_MESSAGE *******************

          SUBROUTINE WRITE_NUMBERS(I1,I2)
          
              WRITE(*,'(I10,A,I10)') I1,'  ---',I2
              WRITE(*,*)
              
          END       

C         *** END WRITE_NUMBERS ***************************

          SUBROUTINE ADD_SIMPLEX
          
          INCLUDE 'triang.cmn'                         
          INTEGER A1,A4,IX(4),AM
              
              A4=MAX0(A(1),A(2),A(3),A(4))
              A1=MIN0(A(1),A(2),A(3),A(4))
              
              DO I=1,4
                 IF (A(I).EQ.A4) IX(4)=I
                 IF (A(I).EQ.A1) IX(1)=I 
              ENDDO
             
              AM=A1
              DO I=1,4
                 IF ((I.NE.IX(1)).AND.
     :               (I.NE.IX(4)).AND.
     :               (A(I).GT.AM)) THEN
                    IX(3)=I 
                    AM=A(I)     
                 ENDIF   
              ENDDO       
      
              DO I=1,4
                 IF ((I.NE.IX(1)).AND.
     :               (I.NE.IX(3)).AND.
     :               (I.NE.IX(4))) IX(2)=I
              ENDDO
                      
              DO I=1,4              
                 SIMP(I,IS)=A(IX(I))
                 IF (IX(I).EQ.4) NEW_AI=I
              ENDDO
              
              OPEN_SIMP(IS)=.TRUE.                         
              
          END 
          
C         *** END ADD_SIMPLEX ********************************
 
C         *****************************************************
C         ***      THIRD LEVE L                             ***
C         ***      FUNCTIONS    -- FACE_EQUAL               ***
C         ***      SUBROUTINES  -- DDC_DECOMPOSITION        ***
C         *****************************************************

          FUNCTION FACE_EQUAL(S1,I1,S2,I2)

              INTEGER S1(4),I1,S2(4),I2
              LOGICAL FACE_EQUAL
              
              FACE_EQUAL=.TRUE.
              J1=0
              J2=0
              
              DO I=1,3
                 J1=J1+1
                 J2=J2+1
                 IF (I.EQ.I1) J1=J1+1
                 IF (I.EQ.I2) J2=J2+1
                 IF (S1(J1).NE.S2(J2)) THEN
                    FACE_EQUAL=.FALSE.
                    GOTO 1
                 ENDIF
              ENDDO
                    
1             CONTINUE
 
          END

C         ****************** END FACE_EQUAL ***************

          SUBROUTINE RANDOM_DISPLACE

          INCLUDE 'triang.cmn'
          COMMON /RCOM/ RND,ISTRAN

              ISTRAN=231 
              J=-1
              JJ=1
 
              DO I=1,NC
                 CALL RANDS
                 RNDX=(RND-0.5D0)*DISP_PAR  
                 CENT(1,I)=CENT(1,I)+RNDX
                 CALL RANDS
                 RNDY=(RND-0.5D0)*DISP_PAR  
                 CENT(2,I)=CENT(2,I)+RNDY
                 CALL RANDS
                 RNDZ=DSQRT(0.25D0-RNDX**2-RNDY**2)*DISP_PAR
                 JJ=JJ*J
                 CENT(3,I)=CENT(3,I)+RNDZ*JJ
              ENDDO
             
              DO I=1,NC      
                 X=CENT(1,I)
                 Y=CENT(2,I)
                 Z=CENT(3,I)
                 CALL IMAGES(X,Y,Z,CELL,RCELL)
                 CENT(1,I)=X
                 CENT(2,I)=Y
                 CENT(3,I)=Z
              ENDDO

              DO I=1,NN
                 IF (IN_CENT(INEIG(I))) THEN
                    NEIG(1,I)=CENT(1,IND_CENT(INEIG(I)))
                    NEIG(2,I)=CENT(2,IND_CENT(INEIG(I)))
                    NEIG(3,I)=CENT(3,IND_CENT(INEIG(I)))
                 ELSE
                    CALL RANDS
                    RNDX=(RND-0.5D0)*DISP_PAR  
                    NEIG(1,I)=NEIG(1,I)+RNDX
                    CALL RANDS
                    RNDY=(RND-0.5D0)*DISP_PAR  
                    NEIG(2,I)=NEIG(2,I)+RNDY
                    RNDZ=DSQRT(0.25D0-RNDX**2-RNDY**2)*DISP_PAR
                    JJ=JJ*J
                    NEIG(3,I)=NEIG(3,I)+RNDZ*real(JJ)
                 ENDIF
              ENDDO
             
              DO I=1,NN                       
                 X=NEIG(1,I)
                 Y=NEIG(2,I)
                 Z=NEIG(3,I)
                 CALL IMAGES(X,Y,Z,CELL,RCELL)
                 NEIG(1,I)=X
                 NEIG(2,I)=Y
                 NEIG(3,I)=Z                 
              ENDDO

              CALL TRIANG_WARNING_MESSAGE
     :        ('degeneration is detected,',
     :         ' subroutine RANDOM_DISPLACE is used')
               
          END

C         *** END RANDOM_DISPLACE **************************** 

          SUBROUTINE SAVE_POSITION

          INCLUDE 'triang.cmn'

          DO I=1,NC
             DO J=1,3
                OLD_CENT(J,I)=CENT(J,I)
             ENDDO
          ENDDO

          DO I=1,NN
             DO J=1,3
                OLD_NEIG(J,I)=NEIG(J,I)
             ENDDO
          ENDDO

        END

C       *** END SAVE_POSITION ************************

        SUBROUTINE RESTORE_POSITION

        INCLUDE 'triang.cmn'

          DO I=1,NC
             DO J=1,3
                CENT(J,I)=OLD_CENT(J,I)
             ENDDO
          ENDDO

          DO I=1,NN
             DO J=1,3
                NEIG(J,I)=OLD_NEIG(J,I)
             ENDDO
          ENDDO
 
        END

C       *** END RESTORE_POSITION ***********************         

         SUBROUTINE SIMPL_TETR
                
         INCLUDE 'triang.cmn'
         DOUBLE PRECISION LG(6),RX(4),RY(4),RZ(4)
                
             DO I=1,IS
 
                DO J=1,4
                   II=SIMP(J,I)
                   IF (IN_CENT(II)) THEN
                      IND=IND_CENT(II)
                      RX(J)=CENT(1,IND)
                      RY(J)=CENT(2,IND)
                      RZ(J)=CENT(3,IND)
                   ELSE IF (IN_NEIG(II)) THEN
                      IND=IND_NEIG(II)
                      RX(J)=NEIG(1,IND)
                      RY(J)=NEIG(2,IND)
                      RZ(J)=NEIG(3,IND)
                   ELSE
                      CALL TRIANG_ERROR_MESSAGE
     :                ('an atom is not centers as well ',
     :                 ' neighbours')
                   ENDIF

                ENDDO
                   
                IL=0   
                GLG=0D0
 
                DO J=1,3
                   DO K=J+1,4
                   
                      IL=IL+1
                      X=RX(J)-RX(K)
                      Y=RY(J)-RY(K)
                      Z=RZ(J)-RZ(K)                      
                      CALL IMAGES(X,Y,Z,CELL,RCELL)
                      
                      LG(IL)=X**2+Y**2+Z**2
                      LG(IL)=DSQRT(LG(IL))
                      GLG=LG(IL)+GLG

                   ENDDO
                ENDDO  
 
                TETR=0D0
                   
                DO J=1,IL-1
                   DO K=J+1,IL
                      TETR=TETR+(LG(J)-LG(K))**2
                   ENDDO
                ENDDO
                      
                GLG=GLG/IL
                SIMP_TET(I)=TETR/(GLG**2)  

             ENDDO
                      
         END
                   
C        *** END SIMPL_TETR *********************
          
        subroutine rands
            
C    mieszany generator liczb losowych o okresie 2**31
 
        INTEGER x
        DOUBLE PRECISION RND
            
        common /RCOM/ rnd,x
 
        integer flag,ax,c,halfm,ax1,ax2,n,ix1,ix2,m
        integer ix3,ix4,ixa,ixq,mic,ixb
        double precision s

c        data flag,halfm,m,n/0,1073741824,65536,32768/
        
        halfm=1073741824
        m=65536
        n=32768         
  
c        if(flag.eq.1) goto 100 

        ax=8*idint(halfm*datan(1.d0)/8.d0)+5
        c=2*idint(halfm*(.5d0-dsqrt(3.d0)/6.d0))+1
        ax1=ax/n
        ax2=ax-ax1*n
        mic=(halfm-c)+halfm
        s=.5d0/halfm
100     ix1=x/m
        ix2=x-ix1*m  
        ixq=ax2*ix2
        ix3=ax1*ix2
        ix4=2*ax2*ix1
        ix3=mod(ix3,m)
        ix4=mod(ix4,m)
        ixa=ixq/n
        ixb=ixq-ixa*n
        ixa=ix3+ix4+ixa 
        ixa=mod(ixa,m)
        x=ixa*n+ixb
        if(x.gt.mic) x=(x-halfm)-halfm
        x=x+c   
        rnd=x*s
        flag=1
        return

      end

